###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import hmac
from everysk.config import settings
from everysk.core.exceptions import SigningError


SEPARATOR: bytes = b':'


def sign(data: str | bytes | bytearray, hash_name: str = 'sha1') -> bytes:
    """
    Sign the data with the SECRET_KEY.
    Possible hash algorithms are 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'.

    Args:
        data (str | bytes | bytearray): The data to be signed.
        hash_name (str): The name of the hash algorithm to be used.
    """
    if isinstance(data, str):
        data = data.encode()

    digest = hmac.new(settings.EVERYSK_SIGNING_KEY.encode(), data, hash_name).hexdigest().encode()
    return SEPARATOR.join([digest, data])


def unsign(signed_data: bytes, hash_name: str = 'sha1') -> bytes | bytearray:
    """
    Unsign the data with the SECRET_KEY.
    Possible hash algorithms are 'md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512'.

    Args:
        signed_data (bytes): The signed data created by the sign method.
        hash_name (str): The name of the hash algorithm to be used.

    Raises:
        SigningError: If the signing key is invalid.
    """
    digest, data = signed_data.split(SEPARATOR, 1)
    if hmac.compare_digest(digest, hmac.new(settings.EVERYSK_SIGNING_KEY.encode(), data, hash_name).hexdigest().encode()):
        return data

    raise SigningError



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from math import ceil
from operator import itemgetter

###############################################################################
#   Globals
###############################################################################
def split_in_slices(total_size: int, slice_size: int) -> list[slice]:
    """
    Divides a range of a given size into slices of approximately equal lengths.

    Args:
        total_size (int): The total size of the range to be divided.
        slice_size (int): The desired size for each slice.

    Returns:
        list[slice]: A list of `slice` objects, where each slice represents
                    a segment of the range. The slices cover the entire range
                    without overlap.

    Examples:
        >>> split_in_slices(1, 2)
        [slice(0, 1)]
        >>> split_in_slices(3, 2)
        [slice(0, 2), slice(2, 3)]
        >>> split_in_slices(5, 3)
        [slice(0, 2), slice(2, 4), slice(4, 5)]
    """
    # Return an empty list if total_size or slice_size is invalid
    if total_size <= 0 or slice_size <= 0:
        return []

    # If the total_size is smaller than or equal to the desired slice size,
    # return a single slice that covers the entire range
    if total_size <= slice_size:
        return [slice(0, total_size)]

    # Calculate the number of slices needed
    n = int(ceil(total_size / float(slice_size)))

    # Compute the base slice size and the remainder
    k, m = divmod(total_size, n)

    # Generate slices using a list comprehension
    return [slice(i * k + min(i, m), (i + 1) * k + min(i + 1, m)) for i in range(n)]

def slices(lst: list, size: int) -> list:
    """
    Split the list in slices.
    https://stackoverflow.com/a/312464

    Args:
        lst (list): The list to split.
        size (int): The size of each slice.

    Returns:
        list: A list in which each element is a slice of the original list.

    Example:
        >>> from everysk.core.lists import slices
        >>> slices([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 2)
        [[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]]

    """
    return [lst[i:i + size] for i in range(0, len(lst), size)]

def sort_list_dict(*keys: list[str], lst: list[dict], reverse: bool = False) -> list:
    """
    Order a list of dictionaries by a specific key or keys.
    https://stackoverflow.com/a/73050

    Args:
        keys (list[str]): The key or keys to sort by.
        lst (list[dict]): The list of dictionaries to sort.
        reverse (bool): Whether to sort in reverse order.

    Returns:
        list: The sorted list of dictionaries.

    Example:
        >>> from everysk.core.lists import sort_list_dict
        >>> sort_list_dict('name', 'age', lst=[{'name': 'John', 'age': 25}, {'name': 'Jane', 'age': 22}], reverse=True)
        [{'name': 'John', 'age': 25}, {'name': 'Jane', 'age': 22}]
    """
    return sorted(lst, key=itemgetter(*keys), reverse=reverse)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
# pylint: disable=protected-access, attribute-defined-outside-init
import pickle
from copy import copy, deepcopy
from typing import Any, Self, Union
from _collections_abc import dict_items, dict_keys, dict_values

from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import DefaultError, FieldValueError, RequiredError
from everysk.core.fields import BoolField
from everysk.core.object import (
    BaseField, BaseDict, BaseObject, BaseObjectConfig,
    BaseDictConfig, _required, _validate, MetaClass, CLASS_KEY
)
from everysk.core.unittests import TestCase


###############################################################################
#   Required Class Test Case
###############################################################################
class RequiredTestCase(TestCase):

    def test_required(self):
        self.assertRaises(RequiredError, _required, attr_name='Test', value=None)
        self.assertRaises(RequiredError, _required, attr_name='Test', value='')
        self.assertRaises(RequiredError, _required, attr_name='Test', value=())
        self.assertRaises(RequiredError, _required, attr_name='Test', value=[])
        self.assertRaises(RequiredError, _required, attr_name='Test', value={})

###############################################################################
#   Validate Class Test Case
###############################################################################
class ValidateTestCase(TestCase):

    def test_none(self):
        _validate(attr_name='test', value=None, attr_type=Date)
        _validate(attr_name='test', value=None, attr_type=DateTime)
        _validate(attr_name='test', value=None, attr_type=dict)
        _validate(attr_name='test', value=None, attr_type=float)
        _validate(attr_name='test', value=None, attr_type=int)
        _validate(attr_name='test', value=None, attr_type=list)
        _validate(attr_name='test', value=None, attr_type=str)

    def test_undefined(self):
        _validate(attr_name='test', value=Undefined, attr_type=Date)
        _validate(attr_name='test', value=Undefined, attr_type=DateTime)
        _validate(attr_name='test', value=Undefined, attr_type=dict)
        _validate(attr_name='test', value=Undefined, attr_type=float)
        _validate(attr_name='test', value=Undefined, attr_type=int)
        _validate(attr_name='test', value=Undefined, attr_type=list)
        _validate(attr_name='test', value=Undefined, attr_type=str)

    def test_date(self):
        _validate(attr_name='test', value=Date.today(), attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=Date)

    def test_datetime(self):
        _validate(attr_name='test', value=DateTime.now(), attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=DateTime)

    def test_dict(self):
        _validate(attr_name='test', value={}, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=dict)

    def test_float(self):
        _validate(attr_name='test', value=1.0, attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=float)

    def test_int(self):
        _validate(attr_name='test', value=1, attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=int)

    def test_list(self):
        _validate(attr_name='test', value=[], attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=list)

    def test_str(self):
        _validate(attr_name='test', value='1', attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=str)

    def test_any(self):
        _validate(attr_name='test', value=None, attr_type=Any)
        _validate(attr_name='test', value={}, attr_type=Any)
        _validate(attr_name='test', value=[], attr_type=Any)
        _validate(attr_name='test', value=123, attr_type=Any)
        _validate(attr_name='test', value='123', attr_type=Any)
        _validate(attr_name='test', value=callable, attr_type=Any)
        _validate(attr_name='test', value=Undefined, attr_type=Any)
        _validate(attr_name='test', value=Undefined.default_parse_string, attr_type=Any)

    def test_callable(self):
        _validate(attr_name='test', value=lambda x: x, attr_type=callable)
        with self.assertRaisesRegex(FieldValueError, 'Key test must be <built-in function callable>.'):
            _validate(attr_name='test', value='func', attr_type=callable)

    def test_annotation_string(self):
        _validate(attr_name='test', value=object(), attr_type='object')
        with self.assertRaisesRegex(FieldValueError, 'Key test must be object.'):
            _validate(attr_name='test', value=str(), attr_type='object')

    def test_generic_alias(self):
        _validate(attr_name='test', value=[], attr_type=list[str])
        with self.assertRaisesRegex(FieldValueError, r'Key test must be list\[str\].'):
            _validate(attr_name='test', value='a', attr_type=list[str])

    def test_union_type(self):
        _validate(attr_name='test', value=1, attr_type=int | str)
        _validate(attr_name='test', value='1', attr_type=int | str)
        with self.assertRaisesRegex(FieldValueError, 'Key test must be int | str.'):
            _validate(attr_name='test', value=1.0, attr_type=int | str)


###############################################################################
#   BaseField Class Test Case
###############################################################################
class BaseFieldTestCase(TestCase):

    def test_init(self):
        field = BaseField(attr_type=str, default='default', required=True, readonly=True, other='Test')
        self.assertEqual(field.attr_type, str)
        self.assertEqual(field.default, 'default')
        self.assertTrue(field.required)
        self.assertTrue(field.readonly)
        self.assertEqual(field.other, 'Test') # pylint: disable=no-member

    def test_init_default_empty(self):
        self.assertRaises(DefaultError, BaseField, attr_type=str, default=[])
        self.assertRaises(DefaultError, BaseField, attr_type=str, default={})

    def test_init_default_not_empty(self):
        BaseField(attr_type=str, default=[1, 2, 3])
        BaseField(attr_type=str, default={'a': 1, 'b': 2})

    def test_init_readonly_default(self):
        self.assertRaises(RequiredError, BaseField, attr_type=str, readonly=True)

    def test_clean(self):
        field = BaseField(attr_type=str)
        self.assertEqual(field.clean_value('Test'), 'Test')

    def test_transform_to_none(self):
        field = BaseField(attr_type=str, empty_is_none=False)
        self.assertEqual(field.transform_to_none(''), '')
        field = BaseField(attr_type=str, empty_is_none=True)
        self.assertEqual(field.transform_to_none(''), None)

    def test_validate_attr_type(self):
        field = BaseField(attr_type=str)
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value=True)
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value='1', attr_type=int)

    def test_validate_readonly(self):
        field = BaseField(attr_type=str, readonly=True, default='Teste')
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value='new')

    def test_validate_required(self):
        field = BaseField(attr_type=str, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)
        field = BaseField(default=Undefined, attr_type=int, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)
        field = BaseField(default=Undefined.default_parse_string, attr_type=int, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)

    def test_validate_required_lazy(self):
        field = BaseField(attr_type=str, required=False, required_lazy=True)
        field.validate(attr_name='test', value=None)
        field = BaseField(attr_type=int, required=False, required_lazy=True)
        field.validate(attr_name='test', value=Undefined)

    def test_repr(self):
        field = BaseField(attr_type=str)
        self.assertEqual(repr(field), 'BaseField')

    def test_get_value(self):
        def func():
            return 'Test'
        field = BaseField(attr_type=callable)
        self.assertEqual(field.get_value(func), 'Test')

    def test_required_and_required_lazy_error(self):
        self.assertRaisesRegex(
            FieldValueError,
            "Required and required_lazy can't be booth True.",
            BaseField,
            attr_type=str,
            required=True,
            required_lazy=True
        )

    def test_equal(self):
        self.assertEqual(
            BaseField(attr_type=str, default='field'),
            BaseField(attr_type=str, default='field')
        )

    def test_not_equal(self):
        self.assertNotEqual(
            BaseField(attr_type=str, default='field1'),
            BaseField(attr_type=str, default='field2')
        )

    def test_getattr_method(self):
        field = BaseField(attr_type=str)
        self.assertTrue(field.startswith)

    def test_getattr_method_error(self):
        field = BaseField(attr_type=str)
        with self.assertRaisesRegex(AttributeError, "type object 'str' has no attribute 'real'"):
            field.real # pylint: disable=pointless-statement

    def test_getattr_method_union_type(self):
        field = BaseField(attr_type=int | str)
        self.assertTrue(hasattr(field, 'startswith'))
        self.assertTrue(hasattr(field, 'real'))

    def test_getattr_method_union_type_error(self):
        field = BaseField(attr_type=int | str)
        with self.assertRaisesRegex(AttributeError, "type object 'int | str' has no attribute 'update'"):
            field.update # pylint: disable=pointless-statement


###############################################################################
#   Fake class for BaseObject tests.
###############################################################################
class BaseObjectTestClass(BaseObject):
    attr: int = None

###############################################################################
#   BaseObjectFields Test Case
###############################################################################
class BaseObjectFieldsTestClass(BaseObject):
    _private: str = 'Private'
    normal_attr = 'Normal'
    field1 = BaseField(attr_type=int)
    field_default = BaseField(attr_type=int, default=100)
    field_readonly = BaseField(attr_type=str, default='default_value', readonly=True)

###############################################################################
#   Fake class for BaseObject tests.
###############################################################################
class BaseObjectAttributeInheritanceTestClass(BaseObjectFieldsTestClass):
    local_attr = BaseField(attr_type=str, default='Local Field')

class BaseObjectInheritanceTestClass(BaseObjectTestClass):
    attr_other: str = None

class BaseObjectRequiredTestClass(BaseObject):
    required = BaseField(attr_type=int, required=True)
    required1 = BaseField(attr_type=int, required=True)

class BaseObjectRequiredInheritanceTestClass(BaseObjectRequiredTestClass):
    required1 = 1

class BaseObjectRequiredLazyTestClass(BaseObject):
    required = BaseField(attr_type=int, required_lazy=True)
    required1 = BaseField(attr_type=int, required_lazy=True)

class BaseObjectRequiredLazyInheritanceTestClass(BaseObjectRequiredLazyTestClass):
    required1 = 1

class BaseObjectAttributeReplace(BaseObject):
    is_parent = BaseField(attr_type=bool, default=True)

class BaseObjectAttributeReplaceChild(BaseObjectAttributeReplace):
    is_parent = False

class BaseFieldClean(BaseField):

    def clean_value(self, value: Any) -> Any:
        if value not in (None, int):
            raise FieldValueError('Error')

class BaseObjectCleanValueError(BaseObject):
    field1 = BaseFieldClean(attr_type=int)

# Fake class for BaseObject
class BaseObjectTestInitClass(BaseObject):
    def __init__(self, **kwargs: dict) -> None:
        raise AttributeError

###############################################################################
#   BaseObject Test Case
###############################################################################
class BaseObjectTestCase(TestCase):

    def test_init(self):
        obj = BaseObjectTestClass()
        self.assertIsNone(obj.attr)

    def test_init_error_handler(self):
        obj = BaseObjectTestInitClass(silent=True)

        self.assertIsInstance(obj, BaseObject)
        self.assertIsNone(obj._errors['after_init'])
        self.assertIsNone(obj._errors['before_init'])
        self.assertIsInstance(obj._errors['init'], AttributeError)

    def test_attr_creation(self):
        obj = BaseObjectTestClass()
        obj.attr = 1
        self.assertEqual(obj.attr, 1)

    def test_fields(self):
        obj = BaseObjectFieldsTestClass()
        self.assertEqual(obj._private, 'Private')
        self.assertEqual(obj.normal_attr, 'Normal')
        self.assertIsNone(obj.field1)
        self.assertEqual(obj.field_default, 100)
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertRaises(FieldValueError, BaseObjectFieldsTestClass, field_readonly='Changed')
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj.field_readonly = 'Changed'

    def test_required(self):
        self.assertRaises(RequiredError, BaseObjectRequiredTestClass)

    def test_required_inheritance(self):
        self.assertRaisesRegex(RequiredError, 'The required attribute is required.', BaseObjectRequiredInheritanceTestClass)
        obj = BaseObjectRequiredInheritanceTestClass(required=2)
        self.assertEqual(obj.required, 2)
        self.assertEqual(obj.required1, 1)

    def test_required_lazy(self):
        obj = BaseObjectRequiredLazyTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )

    def test_required_lazy_inheritance(self):
        obj = BaseObjectRequiredLazyInheritanceTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )
        obj.required = 2
        obj.validate_required_fields()
        self.assertEqual(obj.required, 2)
        self.assertEqual(obj.required1, 1)

    def test_default_receive_none(self):
        obj = BaseObjectFieldsTestClass(field_default=None)
        self.assertEqual(obj.field_default, None)

    def test_attribute_change_bool_type(self):
        parent = BaseObjectAttributeReplace()
        self.assertTrue(parent.is_parent)
        child = BaseObjectAttributeReplaceChild()
        self.assertFalse(child.is_parent)

    def test_get_attribute_class(self):
        self.assertEqual(BaseObjectFieldsTestClass.field_default, 100)

    def test_get_attribute_inheritance_class(self):
        self.assertEqual(BaseObjectAttributeInheritanceTestClass.local_attr, 'Local Field')
        self.assertEqual(BaseObjectAttributeInheritanceTestClass.field_readonly, 'default_value')

    def test_get_attribute_class_error(self):
        with self.assertRaisesRegex(AttributeError, "type object 'BaseObjectFieldsTestClass' has no attribute 'field_default_erro'"):
            BaseObjectFieldsTestClass.field_default_erro # pylint: disable=pointless-statement, no-member

    def test_get_clean_value(self):
        obj = BaseObjectFieldsTestClass()
        self.assertIsNone(obj.field1)
        obj.field1 = 21
        self.assertEqual(obj.field1, 21)

    def test_get_clean_value_error(self):
        with self.assertRaisesRegex(FieldValueError, 'field1: Error'):
            BaseObjectCleanValueError(field1='21')

    def test_del_attribute(self):
        obj = BaseObject(attr1=1)
        self.assertEqual(obj.attr1, 1)
        del obj.attr1
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_del_private_attribute(self):
        obj = BaseObject(_attr1=1)
        self.assertEqual(obj._attr1, 1)
        del obj._attr1
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute '_attr1'"):
            obj._attr1 # pylint: disable=pointless-statement

    def test_copy(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)

    def test_copy_config(self):
        obj1 = BaseObject()
        obj1._config = deepcopy(obj1._config)
        obj1._config.new_field = 'string'
        obj2 = copy(obj1)
        self.assertEqual(BaseObject._config, obj2._config)

    def test_deepcopy(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)

    def test_deepcopy_config(self):
        obj1 = BaseObject()
        obj1._config = deepcopy(obj1._config)
        obj1._config.new_field = 'string'
        obj2 = deepcopy(obj1)
        self.assertEqual(BaseObject._config, obj2._config)

    def test_replace(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)

    def test_replace_config(self):
        obj1 = BaseObject(_private=1, public=2)
        obj1._config = deepcopy(obj1._config)
        obj1._config.new_field = 'string'
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)

    def test_get_state(self):
        obj1 = BaseObject(date=Date(2024, 1, 1))
        obj1._config = deepcopy(obj1._config)
        the_pickle = pickle.loads(pickle.dumps(obj1))

        self.assertEqual(the_pickle.__dict__, {'_errors': None, '_is_frozen': False, '_silent': False, 'date': Date(2024, 1, 1)})

    def test_get_state_raises_key_error_pass(self):
        obj = BaseObject(public={'key': 'value'})
        pickle_element = pickle.loads(pickle.dumps(obj))

        self.assertDictEqual(pickle_element.__dict__['public'], {'key': 'value'})
        self.assertDictEqual(obj.__dict__, {'public': {'key': 'value'}})

    def test_pickle_with_data(self):
        obj = FakeObjectWithAttr(var1="test1", var2="test2")
        ret = pickle.loads(pickle.dumps(obj))

        self.assertIsInstance(obj, type(ret))
        for key in obj.__get_attributes__().keys():
            self.assertTrue(hasattr(ret, key))
            self.assertIsInstance(getattr(obj, key), type(getattr(ret, key)))
            self.assertEqual(getattr(obj, key), getattr(ret, key))

    def test_pickle_empty(self):
        obj = FakeObjectWithAttr()
        ret = pickle.loads(pickle.dumps(obj))

        self.assertIsInstance(obj, type(ret))
        for key in obj.__get_attributes__().keys():
            self.assertTrue(hasattr(ret, key))
            self.assertIsInstance(getattr(obj, key), type(getattr(ret, key)))
            self.assertEqual(getattr(obj, key), getattr(ret, key))

    def test_pickle_private(self):
        obj = FakeObjectWithAttr(_private=1)
        ret = pickle.loads(pickle.dumps(obj))

        self.assertIsInstance(obj, type(ret))
        for key in obj.__get_attributes__().keys():
            self.assertTrue(hasattr(ret, key))
            self.assertIsInstance(getattr(obj, key), type(getattr(ret, key)))
            self.assertEqual(getattr(obj, key), getattr(ret, key))

    def test_to_dict(self):
        obj = BaseObject(_private=1, public=2, obj=BaseObject(attr=1))
        self.assertDictEqual(obj.to_dict(), {'_private': 1, 'public': 2, 'obj': obj.obj})

    def test_to_dict_recursive(self):
        obj = BaseObject(_private=1, public=2, obj=BaseObject(attr=1))
        self.assertDictEqual(obj.to_dict(recursion=True), {'_private': 1, 'public': 2, 'obj': {'attr': 1}})

    def test_to_dict_class(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(
            obj.to_dict(add_class_path=True),
            {
                CLASS_KEY: 'everysk.core.object.BaseDict',
                '_errors': None,
                '_is_frozen': False,
                '_silent': False,
                '_private': 1,
                'public': 2,
                'obj': obj.obj
            }
        )

    def test_to_dict_no_class(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(
            obj.to_dict(add_class_path=False),
            {
                '_private': 1,
                'public': 2,
                'obj': obj.obj
            }
        )

    def test_to_dict_class_defined(self):
        obj = BaseObjectTestClass(attr2=2)
        # The fields with default value does not need to go to the dictionary
        self.assertDictEqual(obj.to_dict(), {'attr2': 2})

    def test_to_dict_exclude_keys(self):
        class TestClass(BaseObject):
            class Config(BaseObjectConfig):
                exclude_keys: set = {'attr'}
            attr: int = 1
            var: str = 2

        obj = TestClass()
        self.assertDictEqual(
            obj.to_dict(add_class_path=True),
            {
                '__class_path__': 'everysk.core._tests.object.TestClass',
                '_errors': None,
                '_is_frozen': False,
                '_silent': False,
                'var': 2
            }
        )

    def test_to_dict_exclude_keys_readonly(self):
        class TestClass(BaseObject):
            attr = BaseField(attr_type=int, default=1, readonly=True)
            var: str = 2

        obj = TestClass()
        self.assertDictEqual(
            obj.to_dict(add_class_path=True),
            {
                '__class_path__': 'everysk.core._tests.object.TestClass',
                '_errors': None,
                '_is_frozen': False,
                '_silent': False,
                'var': 2
            }
        )


    def test_need_validation(self):
        obj = BaseObjectTestClass()
        with self.assertRaisesRegex(FieldValueError, "Key attr must be <class 'int'>."):
            obj.attr = 'a'
        obj._need_validation = False
        obj.attr = 'a'


## Fake class for BaseDict tests.
class BaseDictTestClass(BaseDict):
    attr: int = None
    field = BaseField(attr_type=str)

class BaseDictFieldsTestClass(BaseDict):
    _private: str = 'Private'
    normal_attr = 'Normal'
    field1 = BaseField(attr_type=int)
    field_default = BaseField(attr_type=int, default=100)
    field_readonly = BaseField(attr_type=str, default='default_value', readonly=True)

class BaseDictInheritanceTestClass(BaseDictTestClass):
    attr_other: str = None

class BaseDictRequiredTestClass(BaseDict):
    required = BaseField(attr_type=int, required=True)

class BaseDictRequiredLazyTestClass(BaseDict):
    normal: bool = True
    required = BaseField(attr_type=int, required_lazy=True)

class BaseDictBoolTestClass(BaseDict):
    attr = BoolField()

class FrozenBaseDict(BaseDict):
    class Config(BaseDictConfig):
        keys_blacklist: list = ['invalid']

class BaseDictUndefinedTestClass(BaseDict):
    attr = BaseField(default=Undefined, attr_type=str)

class FakeNPArray:
    def __eq__(self, obj: Any) -> bool:
        raise ValueError('The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()')

class BaseDictNpArrayTestCase(BaseDict):
    attr: FakeNPArray = None

class NpArrayTestCase(TestCase):

    def test_equal_raise(self):
        obj = FakeNPArray()
        with self.assertRaisesRegex(ValueError, r'The truth value of an array with more than one element is ambiguous. Use a.any\(\) or a.all\(\)'):
            assert obj == ''

    def test_set_value(self):
        fake_np_array = FakeNPArray()

        obj = BaseDictNpArrayTestCase()
        obj.attr = fake_np_array
        self.assertIsNotNone(obj.attr)
        self.assertIsNotNone(obj['attr'])

        obj = BaseDictNpArrayTestCase(attr=fake_np_array)
        self.assertIsNotNone(obj.attr)
        self.assertIsNotNone(obj['attr'])

###############################################################################
#   BaseDict Test Case Implementation
###############################################################################
class BaseDictTestCase(TestCase):

    def test_init(self):
        obj = BaseDictTestClass()
        self.assertIsNone(obj.attr)
        self.assertIsNone(obj['field'])

    def test_attr_creation(self):
        obj = BaseDictTestClass()
        obj.attr = 1
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj['attr'], 1)
        obj['attr2'] = 10
        self.assertEqual(obj.attr2, 10) # pylint: disable=no-member
        self.assertEqual(obj['attr2'], 10)

    def test_fields(self):
        obj = BaseDictFieldsTestClass()
        self.assertEqual(obj._private, 'Private')
        self.assertEqual(obj.normal_attr, 'Normal')
        self.assertIsNone(obj.field1)
        self.assertIsNone(obj['field1'])
        self.assertEqual(obj.field_default, 100)
        self.assertEqual(obj['field_default'], 100)
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertEqual(obj['field_readonly'], 'default_value')
        self.assertRaises(FieldValueError, BaseDictFieldsTestClass, field_readonly='Changed')
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj.field_readonly = 'Changed'
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj['field_readonly'] = 'Changed'

    def test_private_attr(self):
        obj = BaseDictFieldsTestClass()
        self.assertNotIn('_private', obj)
        obj._other = 1
        self.assertNotIn('_other', obj)
        self.assertEqual(obj._other, 1)
        with self.assertRaisesRegex(KeyError, "Keys can't start with '_'"):
            obj['_other'] = 1

    def test_update(self):
        obj = BaseDictTestClass()
        self.assertIsNone(obj.attr)
        self.assertIsNone(obj['field'])
        obj.update({'attr': 1, 'field': 'a', 'attr2': 2})
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj['attr'], 1)
        self.assertEqual(obj.attr2, 2) # pylint: disable=no-member
        self.assertEqual(obj['attr2'], 2)
        self.assertRaises(FieldValueError, obj.update, {'attr': '1', 'attr2': 2})

    def test_update_raises_key_error(self):
        with self.assertRaises(KeyError) as context:
            obj = BaseDictTestClass()
            obj.update({'_private': 1})

        self.assertEqual(str(context.exception), '\'The key cannot be called "_private".\'')

    def test_required(self):
        self.assertRaises(RequiredError, BaseDictRequiredTestClass)

    def test_required_lazy(self):
        obj = BaseDictRequiredLazyTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )

    def test_default_receive_none(self):
        obj = BaseDictFieldsTestClass(field_default=None)
        self.assertEqual(obj.field_default, None)
        self.assertEqual(obj['field_default'], None)

    def test_value_persist_on_error(self):
        obj = BaseDictFieldsTestClass()
        try:
            obj['field_readonly'] = 'Changed'
        except FieldValueError:
            pass
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertEqual(obj['field_readonly'], 'default_value')

    def test_clean_value(self):
        base = BaseDictBoolTestClass(attr=1)
        # check clean value changed to bool True
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

        base = BaseDictBoolTestClass()
        base.attr = 1
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

        base = BaseDictBoolTestClass()
        base['attr'] = 1
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

    def test_del_attribute(self):
        obj = BaseDict(attr1=1)
        self.assertEqual(obj['attr1'], 1)
        self.assertEqual(obj.attr1, 1)
        del obj.attr1
        with self.assertRaisesRegex(KeyError, "'attr1'"):
            obj['attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_del_private_attribute(self):
        obj = BaseDict(_attr1=1)
        # Private attributes don't turn into keys
        with self.assertRaisesRegex(KeyError, "'_attr1'"):
            obj['_attr1'] # pylint: disable=pointless-statement

        self.assertEqual(obj._attr1, 1)
        del obj._attr1
        with self.assertRaisesRegex(KeyError, "'_attr1'"):
            obj['_attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute '_attr1'"):
            obj._attr1 # pylint: disable=pointless-statement

    def test_del_key(self):
        obj = BaseDict(attr1=1)
        self.assertEqual(obj['attr1'], 1)
        self.assertEqual(obj.attr1, 1)
        del obj['attr1']
        with self.assertRaisesRegex(KeyError, "'attr1'"):
            obj['attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_get_attribute_raises_error(self):
        with self.assertRaises(AttributeError) as context:
            obj = BaseDict()
            obj.__getattr__('__data__') # pylint: disable=unnecessary-dunder-call

        self.assertEqual(str(context.exception), "'BaseDict' object has no attribute '__data__'.")

    def test_fromkeys(self):
        obj = BaseDict(_private=1, public=2, excluded=3)
        dct = obj.fromkeys(['public', 'new'], 4)
        self.assertEqual(obj._private, dct._private)
        self.assertEqual(obj.public, dct.public)
        self.assertEqual(obj['public'], dct['public'])
        self.assertEqual(dct.new, 4)
        self.assertEqual(dct['new'], 4)

    def test_clear(self):
        # pylint: disable=no-member
        obj = BaseDict(_private=1, public=2)
        self.assertDictEqual(obj.__dict__, {'_private': 1, 'public': 2})
        self.assertEqual(obj._private, 1)
        self.assertEqual(obj.public, 2)
        obj.clear()
        self.assertDictEqual(obj, {})
        self.assertEqual(obj._private, 1)
        self.assertFalse(hasattr(obj, 'public'))

    def test_self_copy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = obj1.copy()
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_copy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_deepcopy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_pop(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(obj.pop('public'), 2)
        self.assertEqual(obj.pop('public', 4), 4)
        with self.assertRaisesRegex(KeyError, 'public'):
            obj.pop('public')

    def test_popitem(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(obj.popitem(), ('public', 2))
        with self.assertRaisesRegex(KeyError, r'popitem\(\)\: dictionary is empty'):
            obj.popitem()

    def test_replace(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertEqual(obj2['public'], 4)

    def test_init_with_undefined(self):
        obj = BaseDictUndefinedTestClass(attr='foo')
        self.assertEqual(obj.attr, 'foo')
        self.assertEqual(obj['attr'], 'foo')
        obj = BaseDictUndefinedTestClass(attr=Undefined.default_parse_string)
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)
        obj = BaseDictUndefinedTestClass(attr=Undefined)
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)
        obj = BaseDictUndefinedTestClass(attr=None)
        self.assertEqual(obj.attr, None)
        self.assertEqual(obj['attr'], None)

    def test_set_undefined(self):
        obj = BaseDictUndefinedTestClass()
        obj.attr = Undefined.default_parse_string
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)

        obj.attr = Undefined
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)

    def test_len(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(len(obj), 1)

    def test_getitem(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(obj['public'], 2)
        self.assertEqual(obj['public'], obj.public)
        self.assertRaisesRegex(KeyError, 'foo', obj.__getitem__, 'foo')
        self.assertRaisesRegex(KeyError, '_private', obj.__getitem__, '_private')

    def test_repr(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(repr(obj), "{'public': 2}")

    def test_or(self):
        # pylint: disable=unnecessary-dunder-call
        obj = BaseDict(_private=1, public=2)

        ret = obj | {'public': 3}
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 1, 'public': 3})

        ret = obj | {'public': 3, 'new': 4}
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 1, 'new': 4, 'public': 3, })

        ret = obj | BaseDict(_private=2, public=3, new=4)
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 2, 'public': 3, 'new': 4})

        self.assertEqual(obj.__or__(1), NotImplemented)

    def test_ror(self):
        # pylint: disable=protected-access, unnecessary-dunder-call
        obj = BaseDict(_private=1, public=2)

        ret = {'public': 3} | obj
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 1, 'public': 2})

        ret = {'public': 3, 'new': 4} | obj
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 1, 'public': 2, 'new': 4})

        ret = obj.__ror__(BaseDict(_private=2, public=3, new=4))
        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret.__dict__, {'_private': 1, 'public': 2, 'new': 4})

        self.assertEqual(obj.__ror__(1), NotImplemented)

    def test_ior(self):
        # pylint: disable=no-member
        obj = BaseDict(_private=1, public=2)

        obj |= {'public': 3}
        self.assertIsInstance(obj, BaseDict)
        self.assertDictEqual(obj.__dict__, {'_private': 1, 'public': 3})

        obj |= {'public': 3, 'new': 4}
        self.assertIsInstance(obj, BaseDict)
        self.assertDictEqual(obj.__dict__, {'_private': 1, 'public': 3, 'new': 4})

        obj |= BaseDict(_private=2, public=3, new=4)
        self.assertIsInstance(obj, BaseDict)
        self.assertDictEqual(obj.__dict__, {'_private': 2, 'public': 3, 'new': 4})

    def test_setstate(self):
        obj = BaseDict(_private=1, public=2)
        ret = pickle.loads(pickle.dumps(obj))

        self.assertIsInstance(ret, BaseDict)
        self.assertDictEqual(ret, obj)
        self.assertDictEqual(
            ret.__dict__,
            {'_errors': None, '_is_frozen': False, '_private': 1, '_silent': False, 'public': 2}
        )

    def test_setitem_raises_error_keys_blacklist(self):
        with self.assertRaises(KeyError) as context:
            a = FrozenBaseDict(invalid=1)
            a['invalid'] = 2

        self.assertEqual(str(context.exception), '\'The key cannot be called "invalid".\'')

    def test_attribute_before_init(self):
        class Test(BaseDict):
            def __init__(self, **kwargs) -> None:
                self.field = 1
                super().__init__(**kwargs)

        obj = Test()
        self.assertEqual(obj.field, 1)

    def test_iteration(self):
        obj = BaseDict(a=1, b=2)
        self.assertListEqual(list(obj), ['a', 'b'])

    def test_need_validation(self):
        obj = BaseDictTestClass()
        with self.assertRaisesRegex(FieldValueError, "Key attr must be <class 'int'>."):
            obj['attr'] = 'a'
        obj._need_validation = False
        obj['attr'] = 'a'

    def test_to_dict(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(obj.to_dict(), {'_private': 1, 'public': 2, 'obj': obj.obj})

    def test_to_dict_recursive(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(obj.to_dict(recursion=True), {'_private': 1, 'public': 2, 'obj': {'attr': 1}})

    def test_to_dict_class(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(
            obj.to_dict(add_class_path=True),
            {
                CLASS_KEY: 'everysk.core.object.BaseDict',
                '_errors': None,
                '_is_frozen': False,
                '_silent': False,
                '_private': 1,
                'public': 2,
                'obj': obj.obj
            }
        )

    def test_to_dict_no_class(self):
        obj = BaseDict(_private=1, public=2, obj=BaseDict(attr=1))
        self.assertDictEqual(
            obj.to_dict(add_class_path=False),
            {
                '_private': 1,
                'public': 2,
                'obj': obj.obj
            }
        )

    def test_to_dict_class_defined(self):
        obj = BaseDictTestClass(attr2=2)
        # The fields with default value that not are privates goes to the dictionary
        self.assertDictEqual(obj.to_dict(), {'attr': None, 'attr2': 2, 'field': None})


class BaseDictProperty(BaseDict):
    _to_date = BaseField(attr_type=str)

    @property
    def to_date(self):
        return self._to_date

    @to_date.setter
    def to_date(self, value):
        self._to_date = value

class BaseDictPropertyInherit(BaseDictProperty):
    pass


class BaseDictPropertyTestCase(TestCase):

    def test_property_setattr_init(self):
        obj = BaseDictProperty(to_date='Test')
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)

    def test_property_setattr(self):
        obj = BaseDictProperty()
        obj.to_date = 'Test'
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)

    def test_property_setitem(self):
        obj = BaseDictProperty()
        obj['to_date'] = 'Test'
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)


class TestBaseObjectConfig(BaseObject):
    class Config:
        value: int = 1

class TestBaseDictConfig(BaseDict):
    class Config(BaseDictConfig):
        value: int = 1

    _config: Config = None

class TestConfigClass1(BaseObject):
    class Config:
        pass

class TestConfigClass2(TestConfigClass1):
    pass

class MetaClassConfigTestCase(TestCase):

    def setUp(self) -> None:
        # pylint: disable=no-member
        TestBaseObjectConfig._config.value = 1
        TestBaseDictConfig._config.value = 1

    def test_object_config_attribute(self):
        # pylint: disable=no-member
        self.assertEqual(TestBaseObjectConfig._config.value, 1)
        obj = TestBaseObjectConfig()
        self.assertEqual(obj._config.value, TestBaseObjectConfig._config.value)

    def test_object_config_singleton(self):
        # pylint: disable=no-member
        obj = TestBaseObjectConfig()
        obj._config.value = 3
        self.assertEqual(TestBaseObjectConfig._config.value, 3)
        self.assertEqual(obj._config.value, TestBaseObjectConfig._config.value)

    def test_object_config_delete(self):
        obj = TestBaseObjectConfig()
        with self.assertRaisesRegex(AttributeError, "type object 'TestBaseObjectConfig' has no attribute 'Config'"):
            TestBaseObjectConfig.Config # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'TestBaseObjectConfig' object has no attribute 'Config'"):
            obj.Config # pylint: disable=pointless-statement

    def test_dict_config_attribute(self):
        self.assertEqual(TestBaseDictConfig._config.value, 1)
        obj = TestBaseDictConfig()
        self.assertEqual(obj._config.value, TestBaseDictConfig._config.value)

    def test_dict_config_singleton(self):
        obj = TestBaseDictConfig()
        obj._config.value = 3
        self.assertEqual(TestBaseDictConfig._config.value, 3)
        self.assertEqual(obj._config.value, TestBaseDictConfig._config.value)

    def test_dict_config_delete(self):
        obj = TestBaseDictConfig()
        with self.assertRaisesRegex(AttributeError, "type object 'TestBaseDictConfig' has no attribute 'Config'"):
            TestBaseDictConfig.Config # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'TestBaseDictConfig' object has no attribute 'Config'"):
            obj.Config # pylint: disable=pointless-statement

    def test_dict_config_no_key(self):
        obj = TestBaseDictConfig()
        with self.assertRaisesRegex(KeyError, 'config'):
            obj['config'] # pylint: disable=pointless-statement

    def test_dict_config_attribute_copy(self):
        obj1 = TestBaseDictConfig()
        obj1._config = copy(obj1._config)
        obj1._config.value = 2
        obj2 = TestBaseDictConfig()
        obj3 = TestBaseDictConfig(**obj1)
        self.assertNotEqual(obj1._config, obj2._config)
        self.assertNotEqual(obj1._config, obj3._config)
        self.assertEqual(obj1._config.value, 2)
        self.assertEqual(obj2._config.value, 1)
        self.assertEqual(obj3._config.value, 1)

    def test_return_self(self):
        self.assertDictEqual(
            MetaClass.__call__.__annotations__,
            {'args': tuple, 'kwargs': dict, 'return': Self}
        )

        self.assertDictEqual(
            MetaClass.__new__.__annotations__,
            {'name': str, 'bases': tuple, 'attrs': dict, 'return': Self}
        )

    def test_config_inheritance(self):
        self.assertNotEqual(TestConfigClass1._config, TestConfigClass2._config)


class FrozenObject(BaseObject):
    class Config:
        frozen: bool = True


class FrozenObjectTestCase(TestCase):
    # pylint: disable=no-member
    def setUp(self) -> None:
        self.message = 'Class everysk.core._tests.object.FrozenObject is frozen and cannot be modified.'

    def test_create(self):
        obj = FrozenObject(attr=1, attr2='Banana')
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj.attr2, 'Banana')

    def test_delete(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            del obj.attr

    def test_create_new_attribute(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr2 = 2

    def test_create_update_attribute(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

    def test_copy(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_deepcopy(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_replace(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertTrue(obj2._is_frozen)


class FrozenDict(BaseDict):
    class Config(BaseDictConfig):
        frozen: bool = True


class FrozenDictTestCase(TestCase):
    def setUp(self) -> None:
        self.message = 'Class everysk.core._tests.object.FrozenDict is frozen and cannot be modified.'

    def test_create(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj.attr2, 'Banana')
        self.assertEqual(obj['attr'], 1)
        self.assertEqual(obj['attr2'], 'Banana')

    def test_delete(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            del obj.attr

        with self.assertRaisesRegex(AttributeError, self.message):
            del obj['attr2']

    def test_create_new_attribute(self):
        obj = FrozenDict()
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

        with self.assertRaisesRegex(AttributeError, self.message):
            obj['attr'] = 2

    def test_create_update_attribute(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

        with self.assertRaisesRegex(AttributeError, self.message):
            obj['attr'] = 2

    def test_clear(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.clear()

    def test_self_copy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = obj1.copy()
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_copy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_deepcopy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_pop(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.pop('public')

    def test_popitem(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.popitem()

    def test_update(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.update({'attr': 1})

    def test_replace(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertEqual(obj2['public'], 4)
        self.assertTrue(obj2._is_frozen)


class FakeBaseObject(BaseObject):
    class Config(BaseObject):
        f1: str = 'FakeBaseObject'
        f2 =  BaseField(attr_type=str, default='FakeBaseObject')

class FakeObjectWithAttr(BaseObject):
    var1: str = None
    var2: str = None

class FakeBaseDict(BaseDict):
    class Config(BaseObject):
        f1: str = 'FakeBaseDict'
        f2 = BaseField(attr_type=str, default='FakeBaseDict')

class Fake01PythonClass:
    class Config:
        pass

class Fake02PythonClass:
    class Config:
        pass

class ConfigHashTestCase(TestCase):
    # https://everysk.atlassian.net/browse/COD-2746

    def test_different_values(self):
        # pylint: disable=no-member
        self.assertEqual(FakeBaseObject._config.f1, 'FakeBaseObject')
        self.assertEqual(FakeBaseObject._config.f2, 'FakeBaseObject')
        self.assertEqual(FakeBaseDict._config.f1, 'FakeBaseDict')
        self.assertEqual(FakeBaseDict._config.f2, 'FakeBaseDict')

    def test_python_qualname(self):
        self.assertEqual(Fake01PythonClass.Config.__name__, 'Config')
        self.assertEqual(Fake01PythonClass.Config.__qualname__, 'Fake01PythonClass.Config')
        self.assertEqual(Fake02PythonClass.Config.__name__, 'Config')
        self.assertEqual(Fake02PythonClass.Config.__qualname__, 'Fake02PythonClass.Config')


class MetaClassParent(BaseObject):
    p01 = BaseField(attr_type=str, default='Parent field')
    p02: str = '1'
    p03 = 1
    p04: float

    @property
    def prop(self):
        return self.p01

    @prop.setter
    def prop(self, value: str) -> None:
        self.p01 = value

    def func(self):
        pass


class MetaClassChild(MetaClassParent):
    p01 = BaseField(attr_type=str, default='Child field')
    c02: str = '1'
    c03 = 1
    c04: float


class MetaClassAttributesTestCase(TestCase):

    def test_class_attributes(self):
        attrs = getattr(MetaClassParent, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Parent field'),
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_class_annotations(self):
        self.assertDictEqual(
            MetaClassParent.__annotations__,
            {
                'p01': str,
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_class_inheritance_attributes(self):
        attrs = getattr(MetaClassChild, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Child field'),
                'p02': str,
                'p03': int,
                'p04': float,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_class_inheritance_annotations(self):
        self.assertDictEqual(
            MetaClassChild.__annotations__,
            {
                'p01': str,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_class_inheritance_default_value(self):
        self.assertEqual(MetaClassChild.p01, 'Child field')
        self.assertEqual(MetaClassChild.p02, '1')
        self.assertEqual(MetaClassChild.p03, 1)
        self.assertIsNone(MetaClassChild.p04)
        self.assertIsNone(MetaClassChild().func())

    def test_class_inheritance_changed_value(self):
        old_parent_p02 = MetaClassParent.p02
        old_child_p02 = MetaClassChild.p02
        MetaClassParent.p02 = 'First'
        self.assertEqual(MetaClassParent.p02, 'First')
        self.assertEqual(MetaClassChild.p02, 'First')
        MetaClassChild.p02 = 'Second'
        self.assertEqual(MetaClassParent.p02, 'First')
        self.assertEqual(MetaClassChild.p02, 'Second')
        MetaClassParent.p02 = old_parent_p02
        MetaClassChild.p02 = old_child_p02

    def test_obj_attributes(self):
        obj = MetaClassParent()
        attrs = getattr(obj, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Parent field'),
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_obj_annotations(self):
        # pylint: disable=no-member
        obj = MetaClassParent()
        self.assertDictEqual(
            obj.__annotations__,
            {
                'p01': str,
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_obj_inheritance_attributes(self):
        obj = MetaClassChild()
        attrs = getattr(obj, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Child field'),
                'p02': str,
                'p03': int,
                'p04': float,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_obj_inheritance_annotations(self):
        # pylint: disable=no-member
        obj = MetaClassChild()
        self.assertDictEqual(
            obj.__annotations__,
            {
                'p01': str,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_obj_inheritance_default_value(self):
        obj = MetaClassChild()
        self.assertEqual(obj.p01, 'Child field')
        self.assertEqual(obj.p02, '1')
        self.assertEqual(obj.p03, 1)
        self.assertIsNone(obj.p04)

    def test_obj_inheritance_changed_value(self):
        parent = MetaClassParent()
        child = MetaClassChild()
        parent.p02 = 'First'
        self.assertEqual(parent.p02, 'First')
        self.assertEqual(child.p02, '1')
        child.p02 = 'Second'
        self.assertEqual(parent.p02, 'First')
        self.assertEqual(child.p02, 'Second')

    def test_property_value(self):
        # https://everysk.atlassian.net/browse/COD-3457
        obj = MetaClassParent(p01='01')
        self.assertEqual(obj.p01, '01')
        self.assertEqual(obj.prop, '01')

        obj = MetaClassParent(prop='02')
        self.assertEqual(obj.p01, '02')
        self.assertEqual(obj.prop, '02')


class BeforeInitTestCase(TestCase):

    class BeforeInitBaseObject(BaseObject):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return {'key': 'before'}

    class BeforeInitBaseDict(BaseDict):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return {'key': 'before'}

    class BeforeInitBaseObjectNone(BaseObject):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return None

    class BeforeInitBaseDictNone(BaseDict):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return None

    class BeforeInitBaseObjectRemoveKey(BaseObject):

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            if 'key' in kwargs:
                del kwargs['key']
            return kwargs

    class BeforeInitBaseDictRemoveKey(BaseDict):

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            if 'key' in kwargs:
                del kwargs['key']
            return kwargs

    def test_base_object(self):
        obj = BeforeInitTestCase.BeforeInitBaseObject(key='test')
        self.assertEqual(obj.key, 'before')

    def test_base_object_none(self):
        obj = BeforeInitTestCase.BeforeInitBaseObjectNone(key='test')
        self.assertEqual(obj.key, 'test')

    def test_base_dict(self):
        obj = BeforeInitTestCase.BeforeInitBaseDict(key='test')
        self.assertEqual(obj.key, 'before')
        self.assertEqual(obj['key'], 'before')

    def test_base_dict_none(self):
        obj = BeforeInitTestCase.BeforeInitBaseDictNone(key='test')
        self.assertEqual(obj.key, 'test')
        self.assertEqual(obj['key'], 'test')

    def test_base_object_remove_key(self):
        obj = BeforeInitTestCase.BeforeInitBaseObjectRemoveKey(key='test')
        with self.assertRaisesRegex(AttributeError, "'BeforeInitBaseObjectRemoveKey' object has no attribute 'key'"):
            obj.key

    def test_base_dict_remove_key(self):
        obj = BeforeInitTestCase.BeforeInitBaseDictRemoveKey(key='test')
        with self.assertRaisesRegex(AttributeError, "'BeforeInitBaseDictRemoveKey' object has no attribute 'key'"):
            obj.key

        with self.assertRaisesRegex(KeyError, 'key'):
            obj['key']


class AfterInitTestCase(TestCase):

    class AfterInitBaseObject(BaseObject):
        key: str = None
        def __after_init__(self) -> None:
            self.key = 'after'

    class AfterInitBaseDict(BaseDict):
        key: str = None
        def __after_init__(self) -> None:
            self.key = 'after'

    def test_after_base_object(self):
        # pylint: disable=no-member
        obj = BaseObject(key='test')
        self.assertEqual(obj.key, 'test')

        obj = AfterInitTestCase.AfterInitBaseObject(key='test')
        self.assertEqual(obj.key, 'after')

    def test_after_base_dict(self):
        obj = BaseDict(key='test')
        self.assertEqual(obj.key, 'test')
        self.assertEqual(obj['key'], 'test')

        obj = AfterInitTestCase.AfterInitBaseDict(key='test')
        self.assertEqual(obj.key, 'after')
        self.assertEqual(obj['key'], 'after')


class SilentTestCase(TestCase):

    class SilentBaseObject(BaseObject):
        f1: bool = False

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            if 'before' in kwargs:
                raise ValueError('before')
            return kwargs

        def __after_init__(self) -> None:
            if hasattr(self, 'after'):
                raise ValueError('after')

        def _init_error_handler(self, kwargs: dict, errors: dict[str, Exception]) -> None:
            for key, error in errors.items():
                if error:
                    self.error_found = f'{key}: {error}'

    class SilentInheritance(SilentBaseObject):
        def __after_init__(self) -> None:
            if hasattr(self, 'after'):
                raise ValueError('after')

    def test_init_error(self):
        self.assertRaisesRegex(FieldValueError, "Key f1 must be <class 'bool'>.", SilentTestCase.SilentBaseObject, f1=2)

    def test_init_inheritance_error(self):
        self.assertRaisesRegex(FieldValueError, "Key f1 must be <class 'bool'>.", SilentTestCase.SilentInheritance, f1=2)

    def test_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(f1=2, silent=True)
        self.assertEqual(obj.error_found, "init: Key f1 must be <class 'bool'>.")

    def test_init_error_inheritance_silent(self):
        obj = SilentTestCase.SilentInheritance(f1=2, silent=True)
        self.assertEqual(obj.error_found, "init: Key f1 must be <class 'bool'>.")

    def test_before_init_error(self):
        self.assertRaisesRegex(ValueError, "before", SilentTestCase.SilentBaseObject, before=2)

    def test_before_init_inheritance_error(self):
        self.assertRaisesRegex(ValueError, "before", SilentTestCase.SilentInheritance, before=2)

    def test_before_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(before=2, silent=True)
        self.assertEqual(obj.error_found, "before_init: before")

    def test_before_init_error_inheritance_silent(self):
        obj = SilentTestCase.SilentInheritance(before=2, silent=True)
        self.assertEqual(obj.error_found, "before_init: before")

    def test_after_init_error(self):
        self.assertRaisesRegex(ValueError, "after", SilentTestCase.SilentBaseObject, after=2)

    def test_after_init_inheritance_error(self):
        self.assertRaisesRegex(ValueError, "after", SilentTestCase.SilentInheritance, after=2)

    def test_after_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(after=2, silent=True)
        self.assertEqual(obj.error_found, "after_init: after")

    def test_after_init_error_inheritance_silent(self):
        obj = SilentTestCase.SilentInheritance(after=2, silent=True)
        self.assertEqual(obj.error_found, "after_init: after")


class TypingCheckingTestCase(TestCase):
    ## Internal classes for testing typing.
    class MultiType(BaseObject):
        field: str | int = None

    class MultiTypeUnion(BaseObject):
        field: Union[str, int] = None

    class SubscriptableType(BaseObject):
        field01: list[int] = None
        field02: tuple[int] = None
        field03: dict[str, int] = None

    ## Testes for typing.
    def test_multi_type_int(self):
        obj = TypingCheckingTestCase.MultiType(field=1)
        self.assertEqual(obj.field, 1)

    def test_multi_type_union_int(self):
        obj = TypingCheckingTestCase.MultiTypeUnion(field=1)
        self.assertEqual(obj.field, 1)

    def test_multi_type_str(self):
        obj = TypingCheckingTestCase.MultiType(field='a')
        self.assertEqual(obj.field, 'a')

    def test_multi_type_union_str(self):
        obj = TypingCheckingTestCase.MultiTypeUnion(field='a')
        self.assertEqual(obj.field, 'a')

    def test_multi_type_error(self):
        with self.assertRaisesRegex(FieldValueError, "Key field must be str | int."):
            TypingCheckingTestCase.MultiType(field=1.0)

    def test_multi_type_union_error(self):
        with self.assertRaisesRegex(FieldValueError, "Key field must be str | int."):
            TypingCheckingTestCase.MultiTypeUnion(field=1.0)

    def test_subscriptable_type_list(self):
        obj = TypingCheckingTestCase.SubscriptableType(field01=[1])
        self.assertListEqual(obj.field01, [1])

    def test_subscriptable_type_tuple(self):
        obj = TypingCheckingTestCase.SubscriptableType(field02=(1,))
        self.assertTupleEqual(obj.field02, (1,))

    def test_subscriptable_type_dict(self):
        obj = TypingCheckingTestCase.SubscriptableType(field03={'a': 1})
        self.assertDictEqual(obj.field03, {'a': 1})

    def test_subscriptable_type_list_error(self):
        with self.assertRaisesRegex(FieldValueError, r'Key field01 must be list\[int\].'):
            TypingCheckingTestCase.SubscriptableType(field01='a')

    def test_subscriptable_type_tuple_error(self):
        with self.assertRaisesRegex(FieldValueError, r'Key field02 must be tuple\[int\].'):
            TypingCheckingTestCase.SubscriptableType(field02='a')

    def test_subscriptable_type_dict_error(self):
        with self.assertRaisesRegex(FieldValueError, r'Key field03 must be dict\[str, int\].'):
            TypingCheckingTestCase.SubscriptableType(field03='a')


class BaseDictSuperTestCase(TestCase):
    # pylint: disable=useless-parent-delegation
    # These tests are only to see if super is supported in BaseDict.
    class Test(BaseDict):

        def clear(self) -> None:
            return super().clear()

        def copy(self) -> dict:
            return super().copy()

        def fromkeys(self, keys: list, default: Any = None) -> dict:
            return super().fromkeys(keys, default)

        def get(self, key: str, default: Any = None) -> Any:
            return super().get(key, default)

        def items(self) -> dict_items:
            return super().items()

        def keys(self) -> dict_keys:
            return super().keys()

        def pop(self, *args) -> Any:
            return super().pop(*args)

        def popitem(self) -> tuple:
            return super().popitem()

        def update(self, *args, **kwargs) -> None:
            return super().update(*args, **kwargs)

        def values(self) -> dict_values:
            return super().values()

    def test_clear(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        obj.clear()
        self.assertEqual(obj, BaseDictSuperTestCase.Test())

    def test_copy(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertDictEqual(obj.copy(), obj)

    def test_fromkeys(self):
        obj = BaseDictSuperTestCase.Test()
        self.assertDictEqual(obj.fromkeys(['a', 'b'], 1), {'a': 1, 'b': 1})

    def test_get(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertEqual(obj.get('a'), 1)
        self.assertEqual(obj.get('b'), None)
        self.assertEqual(obj.get('b', 2), 2)

    def test_items(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertListEqual(list(obj.items()), [('a', 1)])

    def test_keys(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertListEqual(list(obj.keys()), ['a'])

    def test_pop(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertEqual(obj.pop('a'), 1)
        self.assertEqual(obj, BaseDictSuperTestCase.Test())

    def test_popitem(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertEqual(obj.popitem(), ('a', 1))
        self.assertEqual(obj, BaseDictSuperTestCase.Test())

    def test_update(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        obj.update(b=2)
        self.assertEqual(obj, BaseDictSuperTestCase.Test(a=1, b=2))

    def test_values(self):
        obj = BaseDictSuperTestCase.Test(a=1)
        self.assertListEqual(list(obj.values()), [1])



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import json

from everysk.api.api_requestor import APIRequestor
from everysk.core.exceptions import APIError
from everysk.core.unittests import TestCase, mock


###############################################################################
#  API Requestor Test Case Implementation
###############################################################################
class APIRequestorTestCase(TestCase):

    def setUp(self) -> None:
        self.api_requestor = APIRequestor(
            api_entry='https://api.example.com',
            api_version='v2',
            api_sid='your_api_sid',
            api_token='your_api_token',
            verify_ssl_certs=True,
        )
        return super().setUp()

    def test_api_requestor_initialization_without_api_entry_raises_value_error(self):
        with self.assertRaises(ValueError) as context:
            api_requestor = APIRequestor(
            api_entry='',
            api_version='v2',
            api_sid='your_api_sid',
            api_token='your_api_token',
            verify_ssl_certs=True
        )

        self.assertIn("Empty api_entry", str(context.exception))

    def test_api_requestor_initialization_with_wrong_api_version_raises_value_error(self):
        with self.assertRaises(ValueError) as context:
            api_requestor = APIRequestor(
                api_entry='https://api.example.com',
                api_version='v1',
                api_sid='your_api_sid',
                api_token='your_api_token',
                verify_ssl_certs=True
            )

        self.assertIn('Invalid api_version (supported version: "v2")', str(context.exception))

    def test_api_requestor_initialization_without_api_sid_raises_value_error(self):
        with self.assertRaises(ValueError) as context:
            api_requestor = APIRequestor(
                api_entry='https://api.example.com',
                api_version='v2',
                api_sid='',
                api_token='your_api_token',
                verify_ssl_certs=True
            )

        self.assertIn('Invalid api_sid', str(context.exception))

    def test_api_requestor_initialization_without_api_token_raises_value_error(self):
        with self.assertRaises(ValueError) as context:
            api_requestor = APIRequestor(
                api_entry='https://api.example.com',
                api_version='v2',
                api_sid='your_api_sid',
                api_token='',
                verify_ssl_certs=True
            )

        self.assertIn('Invalid api_token', str(context.exception))

    def test_api_requestor_initialization_with_headers_returns_expected_data(self):
        api_requestor = APIRequestor(
            api_entry='https://api.example.com',
            api_version='v2',
            api_sid='your_api_sid',
            api_token='your_api_token',
            verify_ssl_certs=True
        )
        expected_headers = {
            'Accept-Encoding': 'gzip',
            'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
            'Cache-control': 'no-cache',
            'Content-Type': 'application/json',
            'User-Agent': 'Everysk PythonBindings/v2',
            'Authorization': 'Bearer your_api_sid:your_api_token'
        }

        self.assertEqual(api_requestor.headers, expected_headers)

    def test_clean_response_without_expected_status_code_raises_api_error(self):
        api_requestor = APIRequestor(
            api_entry='https://api.example.com',
            api_version='v2',
            api_sid='your_api_sid',
            api_token='your_api_token',
            verify_ssl_certs=True,
        )
        with self.assertRaises(APIError) as context:
            api_requestor._clean_response(400, json.dumps('bad request'))

        self.assertIn('bad request', str(context.exception))

    def test_clean_response_with_expected_status_code_returns_expected_data(self):
        api_requestor = APIRequestor(
            api_entry='https://api.example.com',
            api_version='v2',
            api_sid='your_api_sid',
            api_token='your_api_token',
            verify_ssl_certs=True,
        )
        response = api_requestor._clean_response(200, json.dumps('ok'))

        self.assertEqual(response, 'ok')

    def test_api_requestor_get_method_returns_expected_response(self):
        with mock.patch('everysk.api.http_client.RequestsClient.request') as mock_request:
            mock_request.return_value = (200, json.dumps('ok'))
            response = self.api_requestor.get('/path', {'param': 'value'})

        self.assertEqual(response, 'ok')

    def test_api_requestor_post_method_returns_expected_response(self):
        with mock.patch('everysk.api.http_client.RequestsClient.request') as mock_request:
            mock_request.return_value = (200, json.dumps('ok'))
            response = self.api_requestor.post('/path', {'param': 'value'})

        self.assertEqual(response, 'ok')

    def test_api_requestor_put_method_returns_expected_response(self):
        with mock.patch('everysk.api.http_client.RequestsClient.request') as mock_request:
            mock_request.return_value = (200, json.dumps('ok'))
            response = self.api_requestor.put('/path', {'param': 'value'})

        self.assertEqual(response, 'ok')

    def test_api_requestor_delete_method_returns_expected_response(self):
        with mock.patch('everysk.api.http_client.RequestsClient.request') as mock_request:
            mock_request.return_value = (200, json.dumps('ok'))
            response = self.api_requestor.delete('/path')

        self.assertEqual(response, 'ok')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.datetime import Date, DateTime, ZoneInfo
from everysk.core.exceptions import SDKValueError, FieldValueError, SDKTypeError
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.query import Query


class MockPortfolio(Portfolio):
    _orderable_attributes = ['name', 'date']


###############################################################################
#   Query TestCase Implementation
###############################################################################
class QueryTestCase(TestCase):

    def setUp(self):
        self.mock_klass = MockPortfolio
        self.query = Query(Portfolio)
        self.default_data = {
            'filters': [['name', '=', 'test']],
            'order': ['name'],
            'projection': ['name'],
            'distinct_on': ['date'],
            'limit': 5,
            'offset': 2,
            'page_size': 10,
            'page_token': 'token123'
        }

    def test_init(self):
        # pylint: disable=protected-access
        query = Portfolio.query
        self.assertEqual(query._klass, Portfolio)

        query = Query(Portfolio)
        self.assertEqual(query._klass, Portfolio)

        query = Query(_klass=Portfolio)
        self.assertEqual(query._klass, Portfolio)

        query = Query(_klass='Portfolio')
        self.assertEqual(query._klass, Portfolio)

        self.assertRaisesRegex(
            FieldValueError,
            "The _klass value 'Foo' must be a class or a string with the class name",
            Query,
            _klass='Foo'
        )

        query = Query(Portfolio, _clean_order=['name', 'date'])
        self.assertIsInstance(query._clean_order, set)
        self.assertEqual(query._clean_order, {'name', 'date'})

        self.assertRaisesRegex(
            FieldValueError,
            f'Key _clean_order must be {set}',
            Query,
            Portfolio,
            _clean_order='Foo'
        )

        with mock.patch.object(Query, 'sort_by') as mock_sort_by, \
                mock.patch.object(Query, 'where') as mock_where, \
                mock.patch.object(Query, 'set_projection') as mock_set_projection, \
                mock.patch.object(Query, 'set_distinct_on') as mock_set_distinct_on, \
                mock.patch.object(Query, 'set_limit') as mock_set_limit, \
                mock.patch.object(Query, 'set_offset') as mock_set_offset, \
                mock.patch.object(Query, 'set_page_size') as mock_set_page_size, \
                mock.patch.object(Query, 'set_page_token') as mock_set_page_token:

            Query(Portfolio, **self.default_data)

        # Asserting the methods were called correctly
        mock_sort_by.assert_called_once_with(self.default_data['order'][0])
        mock_where.assert_called_once_with(*self.default_data['filters'][0])
        mock_set_projection.assert_called_once_with(self.default_data['projection'])
        mock_set_distinct_on.assert_called_once_with(self.default_data['distinct_on'])
        mock_set_limit.assert_called_once_with(self.default_data['limit'])
        mock_set_offset.assert_called_once_with(self.default_data['offset'])
        mock_set_page_size.assert_called_once_with(self.default_data['page_size'])
        mock_set_page_token.assert_called_once_with(self.default_data['page_token'])

    def test_init_from_query_dict(self):
        query = Query(Portfolio)
        query.where('name', 'test')
        query.where('tags', ['tag1', 'tag2'])
        query.where('created_on', '2023-08-10')
        query.where('date', '2023-08-10')
        query.sort_by('name')
        query.set_projection(['name', 'date'])
        query.set_distinct_on(['date'])
        query.set_limit(5)
        query.set_offset(2)
        query.set_page_size(10)
        query.set_page_token('token')

        new_query = Query(Portfolio, **query)

        self.assertDictEqual(query, new_query)

    def test_find_or_fail(self):
        self.assertFalse(self.query._find_or_fail) # pylint: disable=protected-access
        self.query.set_find_or_fail(True)
        self.assertTrue(self.query._find_or_fail) # pylint: disable=protected-access

    def test_where_with_invalid_operator(self):
        query = Query(self.mock_klass)
        self.assertRaisesRegex(
            SDKValueError,
            "Invalid operator: invalid_operator for property property",
            query.where,
            'property',
            'invalid_operator',
            'value'
        )

    def test_where_with_operator_for_certain_properties(self):
        query = Query(self.mock_klass)
        self.assertRaisesRegex(
            SDKValueError,
            'Filter by property operator must be \'=\'',
            query.where,
            'property',
            '>',
            'value'
        )

    def test_where_2_arguments(self):
        self.query.where('property', 'value')
        self.assertListEqual(self.query.filters, [('property', '=', 'value')])

    def test_where_3_arguments(self):
        query = Query(self.mock_klass)
        query.where('property', '=', 'value')
        self.assertListEqual(query.filters, [('property', '=', 'value')])

    def test_where_4_arguments(self):
        query = Query(self.mock_klass)
        with self.assertRaises(SDKValueError):
            query.where('property', '=', 'value', 'banana')

    def test_where_invalid_operator(self):
        query = Query(self.mock_klass)
        with self.assertRaises(SDKValueError):
            query.where('property', '<>', 'value')

    def test_where_date_property(self):
        query = Query(self.mock_klass)
        query.where('date', '=', '20230810')
        self.assertListEqual(query.filters, [('date', '>=', DateTime(2023, 8, 10, 0, 0, 0, 0, tzinfo=ZoneInfo('UTC'))), ('date', '<=', DateTime(2023, 8, 10, 23, 59, 59, 999999, tzinfo=ZoneInfo('UTC')))])

    def test_where_date_property_with_hyphen(self):
        query = Query(self.mock_klass)
        query.where('date', '=', '2023-08-10')
        self.assertListEqual(query.filters, [('date', '>=', DateTime(2023, 8, 10, 0, 0, 0, 0, tzinfo=ZoneInfo('UTC'))), ('date', '<=', DateTime(2023, 8, 10, 23, 59, 59, 999999, tzinfo=ZoneInfo('UTC')))])

    def test_where_date_instance_property(self):
        query = Query(self.mock_klass)
        query.where('date', '=', Date(2023, 8, 10))
        self.assertListEqual(query.filters, [('date', '>=', DateTime(2023, 8, 10, 0, 0, 0, 0, tzinfo=ZoneInfo('UTC'))), ('date', '<=', DateTime(2023, 8, 10, 23, 59, 59, 999999, tzinfo=ZoneInfo('UTC')))])

    def test_where_tags_property(self):
        query = Query(self.mock_klass)
        query.where('tags', ['tag1', 'tag2'])
        self.assertListEqual(query.filters, [('tags', '=', 'tag1'), ('tags', '=', 'tag2')])

        query = Query(self.mock_klass)
        query.where('tags', 'tag')
        self.assertListEqual(query.filters, [('tags', '=', 'tag')])

    def test_where_tags_invalid(self):
        with self.assertRaisesRegex(SDKTypeError, 'The tags value must be a string or a list of strings'):
            query = Query(self.mock_klass)
            query.where('tags', {'tag1': 'tag2'})

    def test_where_name_property(self):
        query = Query(self.mock_klass)
        query.where('name', 'value')
        expected_filters = [('name', '<', 'valuf'), ('name', '>=', 'value')]
        self.assertListEqual(query.filters, expected_filters)

    def test_where_name_property_special_case(self):
        query = Query(self.mock_klass)
        query.where('name', 'z')
        expected_filters = [('name', '<', '{'), ('name', '>=', 'z')]
        self.assertListEqual(query.filters, expected_filters)

    def test_sort_by_valid_property(self):
        query = Query(self.mock_klass)
        property_name = 'date'
        result = query.sort_by(property_name)

        self.assertEqual(query.order, [property_name])
        self.assertEqual(query._clean_order, {property_name}) # pylint: disable=protected-access
        self.assertIs(result, query)

    def test_sort_by_invalid_property(self):
        query = Query(self.mock_klass)
        property_name = 'invalid_property'
        with self.assertRaises(ValueError):
            query.sort_by(property_name)

    def test_sort_by_duplicate_property(self):
        query = Query(self.mock_klass)
        property_name = 'date'
        query.order.append(property_name)
        query._clean_order.add(property_name) # pylint: disable=protected-access
        with self.assertRaises(ValueError):
            query.sort_by(property_name)

    def test_set_projection_valid_properties(self):
        properties = ['name', 'date']
        result = self.query.set_projection(properties)

        self.assertEqual(self.query.projection, properties)
        self.assertIs(result, self.query)

    def test_set_projection_only_inverse_projection(self):
        properties = ['-name', '-date']
        result = self.query.set_projection(properties)

        self.assertEqual(self.query.projection, properties)
        self.assertIs(result, self.query)

    def test_set_projection_id_on_custom_index(self):
        from everysk.sdk.entities.custom_index.base import CustomIndex # pylint: disable=import-outside-toplevel
        properties = ['id']
        query = Query(CustomIndex)
        result = query.set_projection(properties)

        self.assertEqual(query.projection, properties)
        self.assertIs(result, query)

    def test_set_projection_id_on_private_security(self):
        from everysk.sdk.entities.private_security.base import PrivateSecurity # pylint: disable=import-outside-toplevel
        properties = ['id']
        query = Query(PrivateSecurity)
        result = query.set_projection(properties)

        self.assertEqual(query.projection, properties)
        self.assertIs(result, query)

    def test_set_projection_mixed_projection(self):
        properties = ['name', '-date']
        with self.assertRaises(ValueError) as context:
            self.query.set_projection(properties)

        self.assertEqual(str(context.exception), 'Projection and Inverse Projection should not be set in the same query')

    def test_set_projection_invalid_properties(self):
        properties = ['invalid_property']
        with self.assertRaises(ValueError) as context:
            self.query.set_projection(properties)

        self.assertEqual(str(context.exception), 'Projection properties does not belongs to Portfolio: invalid_property')

    def test_set_projection_string_property(self):
        property_name = 'name'
        result = self.query.set_projection(property_name)

        self.assertEqual(self.query.projection, [property_name])
        self.assertIs(result, self.query)

    def test_set_distinct_on_string_property(self):
        property_name = 'name'
        result = self.query.set_distinct_on(property_name)

        self.assertEqual(self.query.distinct_on, [property_name])
        self.assertIs(result, self.query)

    def test_set_distinct_on_list_property(self):
        properties = ['name', 'date']
        result = self.query.set_distinct_on(properties)

        self.assertEqual(self.query.distinct_on, properties)
        self.assertIs(result, self.query)

    def test_set_limit(self):
        limit_value = 10
        result = self.query.set_limit(limit_value)

        self.assertEqual(self.query.limit, limit_value)
        self.assertIs(result, self.query)

        self.assertRaisesRegex(
            SDKValueError,
            'The limit value must be greater than or equal to 0.',
            self.query.set_limit,
            -1
        )
        self.assertRaisesRegex(
            SDKValueError,
            'The limit value must be an integer.',
            self.query.set_limit,
            [1]
        )

    def test_set_offset(self):
        offset_value = 5
        result = self.query.set_offset(offset_value)

        self.assertEqual(self.query.offset, offset_value)
        self.assertIs(result, self.query)

        self.assertRaisesRegex(
            SDKValueError,
            'The offset value must be greater than or equal to 0.',
            self.query.set_offset,
            -1
        )

        self.assertRaisesRegex(
            SDKValueError,
            'The offset value must be an integer.',
            self.query.set_offset,
            [1]
        )

    def test_set_page_size(self):
        page_size_value = 20
        result = self.query.set_page_size(page_size_value)

        self.assertEqual(self.query.page_size, page_size_value)
        self.assertIs(result, self.query)

        self.assertRaisesRegex(
            SDKValueError,
            'The page_size value must be greater than or equal to 0.',
            self.query.set_page_size,
            -1
        )

        self.assertRaisesRegex(
            SDKValueError,
            'The page_size value must be an integer.',
            self.query.set_page_size,
            [1]
        )

    def test_set_page_token(self):
        token = 'sample_token'
        result = self.query.set_page_token(token)

        self.assertEqual(self.query.page_token, token)
        self.assertIs(result, self.query)

    def test_load_with_offset(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):
            entity = {'id': Portfolio().generate_id(), 'name': 'Test Portfolio'}
            self.query.get_response = mock.MagicMock(return_value=entity)

            result = self.query.load(offset=1)
        self.assertIsInstance(result, Portfolio)
        self.assertEqual(result.name, 'Test Portfolio')

    def test_load_without_offset(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):
            entity = {'id': Portfolio().generate_id(), 'name': 'Test Portfolio'}
            self.query.get_response = mock.MagicMock(return_value=entity)

            result = self.query.load()
        self.assertIsInstance(result, Portfolio)
        self.assertEqual(result.name, 'Test Portfolio')

    def test_loads_with_limit_and_offset(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):

            entities = [{'id': Portfolio().generate_id(), 'name': 'Test Portfolio 1'}, {'id': Portfolio().generate_id(), 'name': 'Test Portfolio 2'}]
            self.query.get_response = mock.MagicMock(return_value=entities)

            results = self.query.loads(limit=2, offset=1)
        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], Portfolio)
        self.assertEqual(results[0].name, 'Test Portfolio 1')

    def test_loads_without_limit_and_offset(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):
            entities = [{'id': Portfolio().generate_id(), 'name': 'Test Portfolio'}]
            self.query.get_response = mock.MagicMock(return_value=entities)

            results = self.query.loads()
        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], Portfolio)
        self.assertEqual(results[0].name, 'Test Portfolio')

    def test_page_with_page_size_and_token(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):

            response = {
                'entities': [{'id': Portfolio().generate_id(), 'name': 'Test Portfolio'}],
                'next_page_token': 'next_token'
            }
            self.query.get_response = mock.MagicMock(return_value=response)

            result = self.query.page(page_size=10, page_token='sample_token')
        self.assertEqual(len(result['entities']), 1)
        self.assertIsInstance(result['entities'][0], Portfolio)
        self.assertEqual(result['next_page_token'], 'next_token')

    def test_page_without_token(self):
        with mock.patch('everysk.sdk.entities.portfolio.base.Portfolio.generate_id', return_value='port_druyZmYpHsXbgTbC8IMmS4B2Q'):

            response = {
                'entities': [{'id': Portfolio().generate_id(), 'name': 'Test Portfolio'}]
            }
            self.query.get_response = mock.MagicMock(return_value=response)

            result = self.query.page(page_size=10)
        self.assertEqual(len(result['entities']), 1)
        self.assertIsInstance(result['entities'][0], Portfolio)

    def test_pages(self):
        portfolio1 = Portfolio(id='port_druyZmYpHsXbgTbC8IMmS4B2Q', name='Test Portfolio 1')
        portfolio2 = Portfolio(id='port_druyZmYpHsXbgTbC8IMmS4B3Q', name='Test Portfolio 2')

        response1 = {
            'entities': [portfolio1],
            'next_page_token': 'next_token'
        }
        response2 = {
            'entities': [portfolio2]
        }

        self.query.page = mock.MagicMock(side_effect=[response1, response2])

        pages_gen = self.query.pages(page_size=10)
        first_page = next(pages_gen)
        self.assertEqual(len(first_page), 1)
        self.assertIsInstance(first_page[0], Portfolio)
        self.assertEqual(first_page[0].name, 'Test Portfolio 1')

        second_page = next(pages_gen)
        self.assertEqual(len(second_page), 1)
        self.assertIsInstance(second_page[0], Portfolio)
        self.assertEqual(second_page[0].name, 'Test Portfolio 2')

    def test_pages_break_condition(self):
        with mock.patch.object(Query, 'page') as mock_page:
            # Simulate two pages of results. First has next_page_token, second doesn't.
            mock_page.side_effect = [
                {'entities': ['entity1', 'entity2'], 'next_page_token': 'next_token'},
                {'entities': ['entity3'], 'next_page_token': None}
            ]

            pages_generator = self.query.pages(page_size=2)
            first_page = next(pages_generator)
            second_page = next(pages_generator)

            # Assert the expected values are being yielded
            self.assertEqual(first_page, ['entity1', 'entity2'])
            self.assertEqual(second_page, ['entity3'])

            # Assert the generator has exhausted
            with self.assertRaises(StopIteration):
                next(pages_generator)

            # Ensure the mock was called twice
            self.assertEqual(mock_page.call_count, 2)

    def test_fetch_ids(self):
        with mock.patch('everysk.sdk.base.BaseSDK.get_response') as mock_get_response:
            mock_get_response.return_value = ['port_druyZmYpHsXbgTbC8IMmS4B2Q']
            result = self.query.fetch_ids(limit=1, offset=2)

        self.assertListEqual(result, ['port_druyZmYpHsXbgTbC8IMmS4B2Q'])
        mock_get_response.assert_called_once_with(self_obj=self.query, params={'limit': 1, 'offset': 2})

        with mock.patch('everysk.sdk.base.BaseSDK.get_response') as mock_get_response:
            mock_get_response.return_value = []
            result = self.query.fetch_ids()

        self.assertListEqual(result, [])
        mock_get_response.assert_called_once_with(self_obj=self.query, params={'limit': Undefined, 'offset': Undefined})

    def test_fetch_id(self):
        with mock.patch('everysk.sdk.base.BaseSDK.get_response') as mock_get_response:
            mock_get_response.return_value = 'port_druyZmYpHsXbgTbC8IMmS4B2Q'
            result = self.query.fetch_id(offset=2)

        self.assertEqual(result, 'port_druyZmYpHsXbgTbC8IMmS4B2Q')
        mock_get_response.assert_called_once_with(self_obj=self.query, params={'offset': 2})

        with mock.patch('everysk.sdk.base.BaseSDK.get_response') as mock_get_response:
            mock_get_response.return_value = None
            result = self.query.fetch_id()

        self.assertIsNone(result)
        mock_get_response.assert_called_once_with(self_obj=self.query, params={'offset': Undefined})



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.unittests import TestCase


###############################################################################
#   SDK Init Test Case Implementation
###############################################################################
# pylint: disable=import-outside-toplevel
# pylint: disable=no-name-in-module
# pylint: disable=unused-import
class SDKInitTestCase(TestCase):

    def test_get_attribute_imports_all_modules(self):
        from everysk.config import settings
        modules = settings.EVERYSK_SDK_MODULES_PATH.keys()
        self.assertEqual(list(modules), ['WorkerBase'])

    def test_get_attribute_raises_error_with_non_existent_module(self):
        with self.assertRaises(ImportError) as context:
            from everysk.sdk import non_existent_module
        self.assertEqual(str(context.exception), "cannot import name 'non_existent_module' from 'everysk.sdk' (/var/app/src/everysk/sdk/__init__.py)")

    def test_get_attribute_imports_worker_base(self):
        from everysk.sdk.worker_base import WorkerBase as expected
        from everysk.sdk import WorkerBase
        self.assertEqual(expected, WorkerBase)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.exceptions import FieldValueError
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.base_list import EntityList


###############################################################################
#   Entity List TestCase Implementation
###############################################################################
class TestEntityList(TestCase):

    def setUp(self):
        self.entity_list = EntityList()

    def test_initialization_with_attributes(self):
        self.assertIsNone(self.entity_list.min_size)
        self.assertIsNone(self.entity_list.max_size)

    def test_validate_called_on_init(self):
        # Mock _validate to check if it's called during initialization
        with mock.patch.object(EntityList, "_validate", return_value=None) as mock_validate:
            EntityList([1, 2, 3])
            self.assertEqual(mock_validate.call_count, 3)

    def test_validate_called_on_setitem(self):
        with mock.patch.object(EntityList, "_validate", return_value=None) as mock_validate:
            self.entity_list.append(1)
            self.entity_list[0] = 2
            self.assertEqual(mock_validate.call_count, 2)

    def test_validate_called_on_insert(self):
        with mock.patch.object(EntityList, "_validate", return_value=None) as mock_validate:
            self.entity_list.insert(0, 2)
            mock_validate.assert_called_once_with(2)

    def test_validate_called_on_append(self):
        with mock.patch.object(EntityList, "_validate", return_value=None) as mock_validate:
            self.entity_list.append(3)
            mock_validate.assert_called_once_with(3)

    def test_validate_called_on_extend(self):
        with mock.patch.object(EntityList, "_validate", return_value=None) as mock_validate:
            self.entity_list.extend([4, 5])
            # Should be called twice since there are two elements in the iterable
            self.assertEqual(mock_validate.call_count, 2)

    def test_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            self.entity_list.append(1)

    def test_derived_class_with_validation(self):
        class DerivedEntityList(EntityList):
            def _validate(self, value):
                if value < 0:
                    raise ValueError("Value should be non-negative!")
                return value

        derived_list = DerivedEntityList([1])
        self.assertListEqual(derived_list, [1])

        derived_list = DerivedEntityList()
        derived_list.append(1)
        self.assertRaisesRegex(
            ValueError,
            "Value should be non-negative!",
            derived_list.append,
            -1
        )

        self.assertRaisesRegex(
            FieldValueError,
            "Unsupported format.",
            DerivedEntityList,
            [['d']]
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime
from everysk.core.exceptions import RequiredError, FieldValueError
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.datastore.base import Datastore


###############################################################################
#   Datastore TestCase Implementation
###############################################################################
class DatastoreTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'dats_12345678',
            'name': 'My Datastore',
            'description': 'This is a sample datastore.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': '1',
            'data': {'key1': 'value1', 'key2': 'value2'},
            'version': 'v1',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9)
        }
        self.datastore = Datastore(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(Datastore.get_id_prefix(), settings.DATASTORE_ID_PREFIX)

    def test_validate(self):
        expected_data = compress({'class_name': 'Datastore', 'method_name': 'validate', 'self_obj': self.datastore.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        datastore: Datastore = self.datastore.copy()

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            datastore.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_validate_error(self):
        datastore: Datastore = self.datastore.copy()
        datastore.data = DateTime.now()
        with self.assertRaisesRegex(FieldValueError, "Datastore data is not a valid json"):
            datastore.validate()

        datastore.data = None
        with self.assertRaisesRegex(RequiredError, "The data attribute is required"):
            datastore.validate()

    def test_query_load_with_id(self):
        expected_data = compress({'class_name': 'Datastore', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'dats_1234567891011211234567890', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Datastore(id='dats_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_query_load(self):
        datastore = Datastore(link_uid='SampleLinkUID', workspace='SampleWorkspace')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            datastore.load()

        expected_data = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': datastore.to_query().to_dict(add_class_path=True), 'params': {'offset': None}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.object import BaseObject
from everysk.core.string import import_from_string, is_string_object, to_string, normalize_string, normalize_string_to_search, is_isin_string_regex
from everysk.core.unittests import TestCase


class FakeClass(BaseObject):
    pass

class StringTestCase(TestCase):

    def test_import_from_string(self):
        obj = FakeClass()
        cls = import_from_string(obj.get_full_doted_class_path())
        self.assertEqual(cls, FakeClass)

    def test_import_from_string_import_error(self):
        self.assertRaisesRegex(
            ImportError,
            "banana doesn't look like a module path.",
            import_from_string,
            'banana'
        )

    def test_import_from_string__error(self):
        self.assertRaisesRegex(
            ImportError,
            'Module "everysk.core._tests.utils" does not define a "FakeClassError" class.',
            import_from_string,
            'everysk.core._tests.utils.FakeClassError'
        )

    def test_is_string_object(self):
        self.assertTrue(is_string_object('Teste'))
        self.assertTrue(is_string_object("Teste"))
        self.assertFalse(is_string_object({}))
        self.assertFalse(is_string_object(126))

    def test_to_string(self):
        self.assertEqual(to_string('teste'), 'teste')
        self.assertEqual(to_string(123), '123')

    def test_basic_normalization(self):
        self.assertEqual(normalize_string('café'), 'café')  # Note: 'e' and the combining accent as separate characters

    def test_special_characters_remain(self):
        self.assertEqual(normalize_string('!@#$%^&*()'), '!@#$%^&*()')

    def test_spaces_remain(self):
        self.assertEqual(normalize_string('Hello World'), 'Hello World')

    def test_empty_string(self):
        self.assertEqual(normalize_string(''), '')

    def test_no_change(self):
        self.assertEqual(normalize_string('abc'), 'abc')

    def test_normalization_with_combining_characters(self):
        # 'a' followed by a combining tilde
        self.assertEqual(normalize_string('a\u0303'), 'ã')

    def test_normalize_to_search_basic(self):
        self.assertEqual(normalize_string_to_search('   Café  '), 'café')

    def test_valid_isin(self):
        self.assertTrue(is_isin_string_regex('US0378331005'))  # Apple Inc. ISIN

    def test_invalid_isin_wrong_length(self):
        self.assertFalse(is_isin_string_regex('US03783310055'))

    def test_invalid_isin_wrong_characters(self):
        self.assertFalse(is_isin_string_regex('US0@3Z8331005'))

    def test_invalid_isin_starts_with_number(self):
        self.assertFalse(is_isin_string_regex('1US037833100'))

    def test_invalid_isin_starts_with_space(self):
        self.assertFalse(is_isin_string_regex(' US037833100'))

    def test_invalid_isin_ends_with_space(self):
        self.assertFalse(is_isin_string_regex('US037833100 '))

    def test_isin_with_non_string_input(self):
        self.assertFalse(is_isin_string_regex(123456789012))



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# Remember that this code will be running on Cloud Functions environment
# then not all modules can be used/imported.
from logging import getLogger, INFO, StreamHandler, Formatter
from os import getenv
from sys import stdout
from redis import Redis

## Create a Logger object
log = getLogger('firestore-cached-document-write')
log.setLevel(INFO)
log.propagate = False # Don't pass message to others loggers
handler = StreamHandler(stdout)
handler.setLevel(INFO)
handler.setFormatter(Formatter('%(asctime)s - %(levelname)s - %(message)s'))
log.addHandler(handler)


class RedisClient:

    ## Private attributes
    _connection: Redis = None

    @property
    def connection(self) -> Redis:
        """
        We use this property to check if Redis is online
        then returning the working connection.
        """
        try:
            self._connection.ping()
        except Exception: # pylint: disable=broad-exception-caught
            # Create a new connection
            self._connection = Redis(
                host=getenv('REDIS_HOST'),
                port=int(getenv('REDIS_PORT'))
            )

        return self._connection

    def delete(self, key):
        return self.connection.delete(key)


def firestore_cached_document_write(event: dict, context: type) -> None: # pylint: disable=unused-argument
    """
    Triggered by a change to a Firestore document.

    Args:
        event (dict): Event payload -> {'oldValue': {}, 'updateMask': {}, 'value': {}}.
        context (google.cloud.functions.Context): Metadata for the event.
    """
    old_document = event.get('oldValue', {})
    new_document = event.get('value', {})
    client = RedisClient()

    if new_document:
        # Create or Update was triggered
        doc = new_document
    else:
        # Delete was triggered
        doc = old_document

    key = doc.get('fields', {}).get('redis_key', None)
    # We just need to delete the key as the original process
    # will take care of creating it if necessary.
    if key is not None:
        key = key['stringValue']
        client.delete(key)
        log.info('Redis host: %s - %s', getenv('REDIS_HOST'), key)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from unittest import IsolatedAsyncioTestCase
from starlette.testclient import TestClient

from everysk.server.applications import create_application
from everysk.server.endpoints import BaseEndpoint
from everysk.server.responses import Response
from everysk.server.routing import RouteLazy


class FakeEndpoint(BaseEndpoint):

    async def get(self):
        return Response(status_code=200)


class RouteLazyTestCaseAsync(IsolatedAsyncioTestCase):

    def setUp(self) -> None:
        self.app = create_application(routes=[RouteLazy('/', 'everysk.server._tests.routing.FakeEndpoint')])
        self.client = TestClient(self.app)

    async def test_security_headers_middleware(self):
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any
import json

from everysk.config import settings
from everysk.core.exceptions import RequiredError, FieldValueError
from everysk.core.fields import ListField, StrField

from everysk.sdk.entities.base import BaseEntity, ScriptMetaClass
from everysk.sdk.entities.fields import EntityNameField, EntityDescriptionField, EntityLinkUIDField, EntityWorkspaceField, EntityDateTimeField, EntityTagsField
from everysk.sdk.entities.script import Script

###############################################################################
#   Datastore Implementation
###############################################################################
class Datastore(BaseEntity, metaclass=ScriptMetaClass):
    """
    This class represents a datastore entity object and provides methods to validate
    and manage the entity object's data.

    Attributes:
        script (Script): The script object associated with the datastore.
        id (StrField): The unique identifier of the datastore.
        workspace (EntityWorkspaceField): The workspace of the datastore.
        name (EntityNameField): The name of the datastore.
        tags (EntityTagsField): The tags of the datastore.
        description (EntityDescriptionField): The description of the datastore.
        link_uid (EntityLinkUIDField): The link UID of the datastore.
        date (EntityDateTimeField): The date associated with the datastore.
        data (Any): The data associated with the datastore.
        level (StrField): The level of the datastore.

        version (StrField): The version of the datastore.
        created_on (DateTimeField): The created on date of the datastore.
        updated_on (DateTimeField): The updated on date of the datastore.

    Example:
        >>> from everysk.sdk.entities.datastore.base import Datastore
        >>> datastore = Datastore()
        >>> datastore.id = 'dats_12345678'
        >>> datastore.name = 'My Datastore'
        >>> datastore.tags = ['tag1', 'tag2']
        >>> datastore.description = 'This is a sample datastore.'
        >>> datastore.workspace = 'my_workspace'
        >>> datastore.date = DateTime.fromisoformat('20220101')
        >>> datastore.level = '1'
        >>> datastore.data = {'key1': 'value1', 'key2': 'value2'}
        >>> datastore.create()
        >>> print(datastore)
        {
            'id': 'dats_12345678',
            'name': 'My Datastore',
            'description': 'This is a sample datastore.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': '20220101',
            'level': '1',
            'data': {'key1': 'value1', 'key2': 'value2'},
            'version': 'v1',
            'created_on': '2021-01-01T00:00:00.000000Z',
            'updated_on': '2021-01-01T00:00:00.000000Z',
        }
    """
    script: Script = None
    _orderable_attributes = ListField(default=['date', 'created_on', 'updated_on', 'name'], readonly=True)

    id = StrField(regex=settings.DATASTORE_ID_REGEX, max_size=settings.DATASTORE_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

    name = EntityNameField()
    description = EntityDescriptionField()
    tags = EntityTagsField()
    link_uid = EntityLinkUIDField()
    workspace = EntityWorkspaceField()

    date = EntityDateTimeField()
    level = StrField(min_size=1, max_size=settings.DATASTORE_LEVEL_MAX_LENGTH)

    data: Any = None

    def __init__(self, data=None, **kwargs) -> None:
        super().__init__(data=data, **kwargs)

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the datastore id field value.

        Returns:
            str: The prefix of the datastore id field value.

        Example:
            >>> Datastore.get_id_prefix()
            'dats_'

        Notes:
            The prefix is typically used to distinguish datastore IDs from other types of IDs
        """
        return settings.DATASTORE_ID_PREFIX

    def validate_type_data(self) -> bool:
        """
        Validates the 'data' attribute of the Datastore entity to ensure it is not None and contains valid JSON data.

        Raises:
            RequiredError: If the 'data' attribute is None.
            FieldValueError: If the 'data' attribute is not valid JSON.

        Returns:
            bool: True if validation succeeds.

        Example:
            >>> datastore = Datastore()
            >>> datastore.data = {'key1': 'value1', 'key2': 'value2'}
            >>> datastore.validate_type_data()  # Validation succeeds
            True
        """

        if self.data is None:
            raise RequiredError('The data attribute is required')

        try:
            json.dumps(self.data)
        except Exception as exc: # pylint: disable=broad-except
            raise FieldValueError('Datastore data is not a valid json') from exc

        return True

    def validate(self) -> bool:
        """
        This method validates the entity object and raises an exception if it is not
        valid. The validation is performed by calling the `validate` method of each field
        of the entity.

        Args:
            self (Self): The entity object to be validated.

        Raises:
            FieldValueError: If the entity object is not valid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity object is valid, False otherwise.

        Example:
            >>> entity = Entity()
            >>> entity.validate()
            True
        """
        self.validate_type_data()
        return self.get_response(self_obj=self)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import StrField, ListField, IntField, RegexField

###############################################################################
#   Settings Implementation
###############################################################################
WORKFLOW_EXECUTION_ID_PREFIX = StrField(default='wfex_', readonly=True)
WORKFLOW_EXECUTION_ID_LENGTH = IntField(default=25, readonly=True)
WORKFLOW_EXECUTION_ID_REGEX = RegexField(default=r'^wfex_[a-zA-Z0-9]{25}', readonly=True)

WORKFLOW_ID_PREFIX = StrField(default='wrkf_', readonly=True)
WORKFLOW_ID_LENGTH = IntField(default=25, readonly=True)
WORKFLOW_ID_REGEX = RegexField(default=r'^wrkf_[a-zA-Z0-9]{25}', readonly=True)

# legacy workflow execution status
WORKFLOW_EXECUTION_STATUS_COMPLETED = StrField(default='COMPLETED', readonly=True)
WORKFLOW_EXECUTION_STATUS_OK = StrField(default='OK', readonly=True)

# current workflow execution status
WORKFLOW_EXECUTION_STATUS_CANCELED = StrField(default='CANCELED', readonly=True)
WORKFLOW_EXECUTION_STATUS_FAILED = StrField(default='FAILED', readonly=True)
WORKFLOW_EXECUTION_STATUS_PREPARING = StrField(default='PREPARING', readonly=True)
WORKFLOW_EXECUTION_STATUS_RUNNING = StrField(default='RUNNING', readonly=True)
WORKFLOW_EXECUTION_STATUS_SUCCEEDED = StrField(default='SUCCEEDED', readonly=True)
WORKFLOW_EXECUTION_STATUS = ListField(default=['CANCELED', 'COMPLETED', 'FAILED', 'OK', 'PREPARING', 'RUNNING', 'SUCCEEDED'], readonly=True)
WORKFLOW_EXECUTION_STATUS_FINISHED = ListField(default=['CANCELED', 'COMPLETED', 'SUCCEEDED', 'FAILED'], readonly=True)
